from peewee import *
from definitions import DB_PATH

db = SqliteDatabase(DB_PATH)


class BaseModel(Model):
    class Meta:
        database = db


class Team(BaseModel):
    name = CharField(primary_key=True)


class Player(BaseModel):
    name = CharField()


class Bookmaker(BaseModel):
    name = CharField(primary_key=True)


class Match(BaseModel):
    date = DateField()
    home = ForeignKeyField(Team, backref='home_matches')
    away = ForeignKeyField(Team, backref='away_matches')
    league = CharField()
    season = CharField(null = True)
    h_goals = IntegerField(null = True)
    a_goals = IntegerField(null = True)
    h_xg = DoubleField(null = True)
    a_xg = DoubleField(null = True)
    h_shots = IntegerField(null = True)
    a_shots = IntegerField(null = True)
    h_onTarget = IntegerField(null = True)
    a_onTarget = IntegerField(null = True)
    h_deep = IntegerField(null = True)
    a_deep = IntegerField(null = True)
    h_ppda = DoubleField(null = True)
    a_ppda = DoubleField(null = True)
    h_importance = DoubleField(null = True)
    a_importance = DoubleField(null = True)
    h_spi = DoubleField(null = True)
    a_spi = DoubleField(null = True)
    spi_prob_h = DoubleField(null = True)
    spi_prob_a = DoubleField(null = True)
    spi_prob_tie = DoubleField(null = True)
    spi_xg_h = DoubleField(null = True)
    spi_xg_a = DoubleField(null = True)
    spi_non_shot_xg_h = DoubleField(null = True)
    spi_non_shot_xg_a = DoubleField(null = True)
    spi_adj_score_h = DoubleField(null = True)
    spi_adj_score_a = DoubleField(null = True)
    h_proj_score = DoubleField(null = True)
    a_proj_score = DoubleField(null = True)

    class Meta:
        indexes = (
            (("date", "home", "away"), True),
        )


class BookmakerMatchOdds(BaseModel):
    match = ForeignKeyField(Match, backref="odds")
    bookmaker = ForeignKeyField(Bookmaker)
    home = DoubleField()
    draw = DoubleField()
    away = DoubleField()

    class Meta:
        primary_key = CompositeKey("match", "bookmaker")


# date,league_id,league,team1,team2,spi1,spi2,prob1,prob2,probtie,proj_score1,proj_score2,importance1,importance2,score1,score2,xg1,xg2,nsxg1,nsxg2,adj_score1,adj_score2

class PlayerMatchData(BaseModel):
    player = ForeignKeyField(Player)
    match = ForeignKeyField(Match)
    team = ForeignKeyField(Team)
    goals = IntegerField()
    own_goals = IntegerField()
    shots = IntegerField()
    xG = DoubleField()
    time = IntegerField()
    position = CharField()
    yellow_card = IntegerField()
    red_card = IntegerField()
    key_passes = IntegerField()
    assists = IntegerField()
    xA = DoubleField()
    xGChain = DoubleField()
    xGBuildup = DoubleField()
    positionOrder = IntegerField()

    class Meta:
        indexes = (
            (("player", "match", "team"), True),
        )


def create_db():
    db.connect()
    db.drop_tables([Match, Team, Player, PlayerMatchData])
    db.create_tables([Match, Team, Player, PlayerMatchData])
    db.close()

