import database.dbmodels as db
import pandas
import numpy as np

from util.util import get_season


def avg_n_last_team_stat(team, home_stat_name, away_stat_name, season, before_date, n):
    matches = db.Match.select().limit(n).where(
        (db.Match.season == season) &
        ((db.Match.home == team.name) | (db.Match.away == team.name)) &
        (db.Match.date < before_date)).order_by(db.Match.date)
    if len(matches) == 0:
        return 0
    stat_sum = 0.0
    for match in matches:
        stat_sum += getattr(match, home_stat_name) if match.home == team else getattr(match, away_stat_name)
    return stat_sum / len(matches)


def get_team_xgA(team, date, season=None, n=None):
    matches = []
    if n is not None:
        matches = db.Match.select().limit(n).where(
            ((db.Match.home == team.name) | (db.Match.away == team.name)) &
            (db.Match.date < date) &
            (db.Match.season == get_season(date)) &
            (db.Match.a_xg.is_null(False) & (db.Match.h_xg.is_null(False)))).order_by(db.Match.date)
    elif season is not None:
        matches = db.Match.select().where(
            ((db.Match.home == team.name) | (db.Match.away == team.name)) &
             (db.Match.date < date) &
            (db.Match.season == season) &
            (db.Match.a_xg.is_null(False) & (db.Match.h_xg.is_null(False)))).order_by(db.Match.date)


    if len(matches) == 0:
        return 0
    stat_sum = 0.0
    for match in matches:
        stat_sum += getattr(match, "a_xg") if match.home == team else getattr(match, "h_xg")
    return stat_sum / len(matches)


def avg_n_last_player_stats(for_match, stat_names):
    sql = '''
        select pmd.player_id, pmd.team_id, {stats}
        from playermatchdata pmd left outer join
          (
            select prev.player_id, {stats_subquery}
            from playermatchdata prev join match m on prev.match_id = m.id
            where m.season == {season} and m.date < \"{date}\"
            group by prev.player_id
          ) prev
            on pmd.player_id == prev.player_id
        where pmd.match_id == {match_id} and pmd.position != "Sub"
        group by pmd.player_id
        order by pmd.team_id, pmd.positionOrder
    '''
    data = db.db.execute_sql(sql.format(
        stats=str.join(", ", ["CASE WHEN avg_" + stat + " IS NOT NULL THEN avg_" + stat + " ELSE 0 END" for stat in
                              stat_names]),
        stats_subquery=str.join(", ", ["avg(prev." + stat + ") as avg_" + stat for stat in stat_names]),
        season=for_match.season,
        date=for_match.date,
        match_id=for_match.id
    ))
    return list(data)


def get_n_last_stat_player_teamlineup(match, team, n, stat, list=True):
    data = []
    players_data = db.PlayerMatchData.select().where(
        (db.PlayerMatchData.match == match) &
        (db.PlayerMatchData.team == team) &
        (db.PlayerMatchData.position != "Sub")
    ).order_by(db.PlayerMatchData.positionOrder)
    for player in players_data:
        player_stat_sum = 0
        recent_matches = db.PlayerMatchData.select().join(db.Match).where(
            (db.PlayerMatchData.player == player.player) &
            (db.Match.date < match.date)
        ).order_by(db.Match.date.desc()).limit(n)
        for recent_player_match in recent_matches:
            player_stat_sum += getattr(recent_player_match, stat)
        if player_stat_sum == 0:
            data.append(player_stat_sum)
        else:
            data.append(player_stat_sum / min(len(recent_matches), n))
    if list:
        return data
    else:
        return sum(data)


def get_season_stat_player_teamlineup(match, team, stat, list=True):
    data = []
    players_data = db.PlayerMatchData.select().where(
        (db.PlayerMatchData.match == match) &
        (db.PlayerMatchData.team == team) &
        (db.PlayerMatchData.position != "Sub")
    ).order_by(db.PlayerMatchData.positionOrder)
    if len(players_data) == 0:
        return 0
    for player in players_data:
        recent_matches = db.PlayerMatchData.select().join(db.Match).where(
            (db.PlayerMatchData.player == player.player) &
            (db.Match.season == match.season) &
             (db.Match.date < match.date)
        ).order_by(db.Match.date.desc())
        player_score = 0
        for recent_player_match in recent_matches:
            player_score += getattr(recent_player_match, stat)
        if recent_matches:
            data.append(player_score/len(recent_matches))
        else:
            data.append(player_score)
    if list:
        return data
    else:
        return sum(data)


def days_since_last(team, date):
    match = db.Match.select().where(
        ((db.Match.home == team) |
        (db.Match.away == team)) &
        (db.Match.date < date)
    ).order_by(db.Match.date.desc()).limit(1)
    if match:
        match = match[0]
        return (date - match.date).days
    return 10000


def count_num_players(match, team):
    players_data = db.PlayerMatchData.select().where(
        (db.PlayerMatchData.match == match) &
        (db.PlayerMatchData.team == team) &
        (db.PlayerMatchData.position != "Sub")
    ).order_by(db.PlayerMatchData.positionOrder)
    if len(players_data) ==11:
        return True
    else:
        return False


def generate_peewee_dataset(leagues, seasons, n_average):
    matches = []
    for season in seasons:
        for league in leagues:
            m = db.Match.select().where(
                (db.Match.league == league) &
                (db.Match.season == season) &
                (db.Match.h_xg.is_null(False)) &
                (db.Match.a_xg.is_null(False))
            ).order_by(db.Match.date.asc())
            if league == "Bundesliga":
                teams_in_league = 18
            elif league == "RFPL":
                teams_in_league = 16
            else:
                teams_in_league = 20
            m = list(m)[(teams_in_league // 2) * 2:]
            matches +=m
    match_data = []
    for match in matches:
        print("{} - {} ({})".format(match.home.name, match.away.name, match.season))
        data = {
            "id": match.get_id,
            "date": match.date,
            "home": match.home.name,
            "season": match.season,
            "league": match.league,
            "h_goals": match.h_goals,
            "a_goals": match.a_goals,
            "away": match.away.name,
            "h_match_xg": match.h_xg,
            "a_match_xg": match.a_xg,
            "h_spi": match.h_spi,
            "a_spi": match.a_spi,
            "h_importance": match.h_importance if match.h_importance else 0,
            "a_importance": match.a_importance if match.a_importance else 0,
            "h_days_since_last": days_since_last(match.home, match.date),
            "a_days_since_last": days_since_last(match.away, match.date),
            "h_days_since_last_exp":  np.math.exp((-days_since_last(match.home, match.date)/7)),
            "a_days_since_last_exp": np.math.exp((-days_since_last(match.away, match.date)/7)),
            "h_recent_team_xG_lineup": get_n_last_stat_player_teamlineup(match, match.home, n_average, "xG", list=False),
            "a_recent_team_xG_lineup": get_n_last_stat_player_teamlineup(match, match.away, n_average, "xG", list=False),
            "h_season_team_xG_lineup": get_season_stat_player_teamlineup(match, match.home, "xG", list=False),
            "a_season_team_xG_lineup": get_season_stat_player_teamlineup(match, match.away, "xG", list=False),
            "h_recent_team_xG_lineup_list": get_n_last_stat_player_teamlineup(match, match.home, n_average, "xG"),
            "a_recent_team_xG_lineup_list": get_n_last_stat_player_teamlineup(match, match.away, n_average, "xG"),
            "h_season_team_xG_lineup_list": get_season_stat_player_teamlineup(match, match.home, "xG"),
            "a_season_team_xG_lineup_list": get_season_stat_player_teamlineup(match, match.away, "xG"),
            "h_recent_team_xGA": get_team_xgA(match.home, match.date, n=n_average),
            "a_recent_team_xGA": get_team_xgA(match.away, match.date, n=n_average),
            "h_season_team_xGA": get_team_xgA(match.home, match.date, season=match.season),
            "a_season_team_xGA": get_team_xgA(match.away, match.date, season=match.season),
            "result": one_hot_result(match.h_goals, match.a_goals)
        }
        if data['h_season_team_xG_lineup'] == 0:
            print("h_season_xG 0")
        match_data.append(data)
    return match_data


def generate_dataframe_dataset(leagues=None, seasons=None, complete_dataset=False):
    if complete_dataset:
        matches = db.Match.select().where((db.Match.h_importance.is_null(False) | db.Match.a_importance.is_null(False))
                                          & (db.Match.h_spi.is_null(False) | db.Match.a_spi.is_null(False)))
    else:
        matches = db.Match.select().where(db.Match.league.in_(leagues)
                                          & db.Match.season.in_(seasons)
                                          & (db.Match.h_importance.is_null(False) | db.Match.a_importance.is_null(False)))
    match_df_all_features = pandas.DataFrame(list(matches.dicts()))
    wma_features = create_wma_features(match_df_all_features, 55)
    match_df = map_features_for_teams_to_single_row(wma_features)
    match_df = pandas.merge(left=match_df_all_features, left_index=True, right=match_df, right_index=True,how='inner')
    match_df = match_df.drop(columns=['away', 'home', 'id_y']).rename(columns={'id_x': 'id'})
    match_df['h_days_since_last'] = match_df.apply(lambda row: days_since_last(row['HomeTeam'], row['date']), axis=1)
    match_df['a_days_since_last'] = match_df.apply(lambda row: days_since_last(row['AwayTeam'], row['date']), axis=1)
    match_df['h_days_since_last_exp'] = match_df.apply(lambda row: np.math.exp((-row['h_days_since_last']/7)), axis=1)
    match_df['a_days_since_last_exp'] = match_df.apply(lambda row:  np.math.exp((-row['a_days_since_last']/7)), axis=1)
    match_df_copy = match_df.copy()
    match_df_copy = normalize_dataframe(match_df_copy.drop(columns=['HomeTeam', 'id', 'AwayTeam', 'league', 'season', 'date',
                                                                    'a_goals','h_goals', 'spi_prob_a', 'spi_prob_tie', 'spi_prob_h',  'h_days_since_last','a_days_since_last',
                                                                    'h_days_since_last_exp','a_days_since_last_exp']))  # Normalize all features.
    match_df.update(match_df_copy)

    match_df['result'] = match_df.apply(lambda row: append_result_value(row), axis=1)
    return match_df


def append_result_value(row):
    return one_hot_result(row['h_goals'], row['a_goals'])


def one_hot_result(home_goals, away_goals):
    if home_goals > away_goals:
        return [1, 0, 0]
    elif away_goals > home_goals:
        return [0, 0, 1]
    else:
        return [0, 1, 0]


def result_label(home_goals, away_goals):
    if home_goals > away_goals:
        return 0
    elif away_goals > home_goals:
        return 2
    else:
        return 1


def normalize_dataframe(dataframe):
    return (dataframe - dataframe.min()) / (dataframe.max() - dataframe.min())


def map_stats_to_single_team_per_row(old_stats_df):
    # Create a list of columns we want and their mappings to more interpretable names
    home_stats_cols = ['home', 'h_goals', 'a_goals', 'h_shots', 'a_shots', 'h_onTarget',
                       'a_onTarget', 'h_spi', 'h_xg', 'a_xg',
                       'spi_xg_h', 'spi_xg_a', 'h_deep', 'a_deep', 'h_ppda', 'a_ppda']

    away_stats_cols = ['away', 'a_goals', 'h_goals', 'a_shots', 'h_shots', 'a_onTarget',
                       'h_onTarget', 'a_spi', 'a_xg', 'h_xg',
                       'spi_xg_a', 'spi_xg_h', 'h_deep', 'a_deep', 'a_ppda', 'h_ppda']

    stats_cols_mapping = ['team', 'goalsFor', 'goalsAgainst', 'shotsFor',
                          'shotsAgainst', 'shotsOnTargetFor', 'shotsOnTargetAgainst',
                          'spiFor', 'xgFor', 'xgAgainst', 'spiXgFor', 'spiXgAgainst', 'deepFor',
                          'deepAgainst','ppdaFor', 'ppdaAgainst']

    # Create a dictionary of the old column names to new column names
    home_mapping = {old_col: new_col for old_col, new_col in zip(home_stats_cols, stats_cols_mapping)}
    away_mapping = {old_col: new_col for old_col, new_col in zip(away_stats_cols, stats_cols_mapping)}

    # Put each team onto an individual row
    multi_line_stats = (old_stats_df[['id'] + home_stats_cols]  # Filter for only the home team columns
                        .rename(columns=home_mapping)  # Rename the columns
                        .assign(homeGame=1)  # Assign homeGame=1 so that we can use a general function later
                        .append((old_stats_df[['id'] + away_stats_cols])  # Append the away team columns
                                .rename(columns=away_mapping)  # Rename the away team columns
                                .assign(homeGame=0), sort=True)
                        .sort_values(by='id')  # Sort the values
                        .reset_index(drop=True))
    return multi_line_stats


def create_wma_features(stats, span):
    # Create a restructured DataFrames so that we can calculate EMA
    multi_line_stats = map_stats_to_single_team_per_row(stats)

    # Create a copy of the DataFrame
    ema_features = multi_line_stats[['id', 'team', 'homeGame']].copy()

    # Get the columns that we want to create EMA for
    feature_names = multi_line_stats.drop(columns=['id', 'team', 'homeGame']).columns

    # Loop over the features
    for feature_name in feature_names:
        feature_ema = (multi_line_stats.groupby('team')[feature_name]  # Calculate the EMA
                       .transform(lambda row: row.ewm(span=span, min_periods=2)
                                  .mean()
                                  .shift(1)))  # Shift the data down 1 so we don't leak data
        ema_features[feature_name] = feature_ema  # Add the new feature to the DataFrame
    return ema_features


def map_features_for_teams_to_single_row(stats_features):
    non_features = ['homeGame', 'team', 'id']
    stats_features_restructured = (stats_features.query('homeGame == 1')
                                   .rename(columns={col: 'wma_' + col + '_Home' for col in stats_features.columns if col not in non_features})
                                   .rename(columns={'team': 'HomeTeam'})
                                   .pipe(pandas.merge, (stats_features.query('homeGame == 0')
                                                    .rename(columns={'team': 'AwayTeam'})
                                                    .rename(columns={col: 'wma_' + col + '_Away' for col in stats_features.columns
                                                                     if col not in non_features})), on=['id']))
    return stats_features_restructured



# matches = db.Match.select().where((db.Match.spi_prob_a.is_null(False) & db.Match.spi_prob_tie.is_null(False)))
# match_df_all_features = pandas.DataFrame(list(matches.dicts()))
#
# num_correct = 0
#
# for idx, match in match_df_all_features[20:].iterrows():
#
#     prediction = [match['spi_prob_h'], match['spi_prob_tie'], match['spi_prob_a']]
#     print("{home} {home_score} - {away_score} {away}: {result}. Prediction: {pred}".format(
#         home = match["home"],
#         away = match["away"],
#         result = one_hot_result(match['h_goals'], match['a_goals']),
#         pred = prediction,
#         home_score = match["h_goals"],
#         away_score = match["a_goals"]
#     ))
#
#     if np.argmax(prediction) == one_hot_result(match['h_goals'], match['a_goals']).index(max(one_hot_result(match['h_goals'], match['a_goals']))):
#         print("Correct prediction")
#         num_correct += 1
#     else:
#         print("Wrong prediction")
#
# print("Total accuracy of test-set: {}".format(str(num_correct/len(match_df_all_features))))