import pickle

import numpy as np
from keras.engine.saving import load_model
from keras.utils import plot_model

from betting import plotting
from betting.betting_evaluator import evaluate_games_weekly
from betting.strategy import FixedBet, FixedReturn, KellyRatio, VarianceAdjusted
from datasets import dataloading
from datasets import dataset_generation as dsg
from datasets.dataloading import separate_test_data
from models import model
from util.model_input_util import create_features_multiple_input, create_multiple_target_output


class ExpectedLineUp(model.KerasModel):

    def __init__(self, network_dims, input_size, model_number, dir_name="expected-lineup", loss="categorical_crossentropy",
                 activation_function="relu"):
        super().__init__(network_dims, input_size, model_number=model_number, activation_function=activation_function, learning_rate=0.0001,
                         dir_name=dir_name, optimizer="adam", loss=loss, batch_size=256)
        plot_model(self.model, "exp.png")

    @staticmethod
    def get_default_features():
        return {
            "feature_list": [
                'h_recent_team_xG_lineup_list',
                'a_recent_team_xG_lineup_list',
                'h_season_team_xG_lineup_list',
                'a_season_team_xG_lineup_list',
                'h_recent_team_xGA',
                'a_recent_team_xGA',
                'h_season_team_xGA',
                'a_season_team_xGA',
                'h_days_since_last_exp',
                'a_days_since_last_exp'
            ]
        }